import os
from dataclasses import field, dataclass

from py2lispIDyOM.configuration import get_timestamp, IDyOMConfiguration, ExperimentLogger


@dataclass
class IDyOMExperiment:
    """
    A class to configure the IDyOM experiment.

    :param test_dataset_path: The path to your test dataset (required)
    :param pretrain_dataset_path: The path to your pretrain dataset
    :param experiment_history_folder_path: The path to which you want to save all the result data/plots, defaults to None.

    """

    test_dataset_path: str
    pretrain_dataset_path: str = None
    experiment_history_folder_path: str = None
    idyom_config: IDyOMConfiguration = field(default_factory=IDyOMConfiguration)

    def __post_init__(self):
        self.logger = ExperimentLogger(pretrain_dataset_path=self.pretrain_dataset_path,
                                       test_dataset_path=self.test_dataset_path,
                                       experiment_history_folder_path=self.experiment_history_folder_path)

    def _update_idyom_config(self):
        test_dataset_id = self._generate_test_dataset_id()
        train_dataset_id = self._generate_train_dataset_id()
        self.idyom_config.run_model_configuration.required_parameters.dataset_id = test_dataset_id
        self.idyom_config.run_model_configuration.training_parameters.pretraining_id = train_dataset_id

        self.idyom_config.database_configuration.this_exp_log_path = self.logger.this_exp_folder
        self.idyom_config.database_configuration.test_dataset_id = test_dataset_id
        self.idyom_config.database_configuration.pretrain_dataset_id = train_dataset_id
        self.idyom_config.run_model_configuration.output_parameters.output_path = self.logger.output_data_exp_folder

    @staticmethod
    def _generate_test_dataset_id() -> str:
        moment = get_timestamp()
        dataset_id = '66' + moment
        return dataset_id

    def _generate_train_dataset_id(self):
        # only generate an ID if pretrain_dataset_path is not None
        if self.pretrain_dataset_path:
            moment = get_timestamp()
            dataset_id = '99' + moment
            return dataset_id
        else:
            pass

    def set_parameters(self, **kwargs):
        """
        Set the IDyOM model parameters.
        :param kwargs: see the README or run_idyom_tutorial for a complete list of parameters (keyword arguments).

        """
        configuration = self.idyom_config.run_model_configuration
        surface_dict: dict = configuration.get_surface_dict()
        # print(f'{surface_dict=}')
        kw2hide_in_errormsg = ['output_path', 'dataset_id', 'pretraining_id', 'stmo_options', 'ltmo_options']
        kw2show = list(surface_dict.keys())
        kw2show = [ele for ele in kw2show if ele not in kw2hide_in_errormsg]
        for key, value in kwargs.items():
            if key not in surface_dict:
                raise KeyError(f'parameter \'{key}\' is invalid. Valid parameters are: {kw2show}')
            configuration.recursive_set_attr(key=key, value=value)

    def _generate_lisp_commands(self):
        """
        Generate the LISP commands for the IDyOM model configurations.
        :return: the total lisp commands for the idyom experiment
        """
        self._update_idyom_config()
        lisp_command = self.idyom_config.to_lisp_command()
        return lisp_command

    def generate_lisp_script(self, write=True):
        """
        Generate the LISP script for the IDyOM model configurations.
        :param write: whether to write the file or not, defaults to True.
        :type write: bool
        :return: the path to the lisp script file.
        """
        self._update_idyom_config()
        path_to_file = self.logger.this_exp_folder
        lisp_file_path = path_to_file + 'compute.lisp'
        lisp_command = self.idyom_config.to_lisp_command()
        if write:
            with open(lisp_file_path, "w") as f:
                f.write(lisp_command)
        return str(lisp_file_path)

    def run(self):
        """
        Run the IDyOM model.
        """

        run_condition = all([
            self.idyom_config.run_model_configuration.required_parameters.is_complete()
        ])
        assert run_condition
        print('** running lisp script **')
        os.system("sbcl --noinform --load " + self.generate_lisp_script())
        print(' ')
        print('** Finished! **')



import os
from typing import List

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from py2lispIDyOM.extract import MelodyInfo, ExperimentInfo


class BasicAxsGeneration:

    @staticmethod
    def generic_idyom_output_along_time(ax: matplotlib.axes.Axes,
                                        melody_info: MelodyInfo,
                                        selected_idyom_output: str,
                                        grid: bool,
                                        color=None):
        valid_idyom_output_keyword_list = melody_info.get_idyom_output_keyword_list()
        if selected_idyom_output in valid_idyom_output_keyword_list:
            selected_output_values = melody_info.get_idyom_output_nparray(selected_idyom_output)
        else:
            raise ValueError(f'selected_idyom_output \'{selected_idyom_output}\' is invalid. '
                             f'Valid outputs are: {valid_idyom_output_keyword_list}')

        onset_in_beat = melody_info._get_onset_beat_nparray()

        ax.plot(onset_in_beat, selected_output_values, "-o", color=color)

        ax.margins(x=0.01)
        ax.margins(y=0)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        xticks = np.arange(0, np.amax(onset_in_beat))
        ax.set_xticks(ticks=xticks, minor=True)
        ax.set_xlabel('Time in quarter note')
        yticks = np.arange(0, np.amax(selected_output_values) + 1)
        ax.set_yticks(ticks=yticks, minor=True)
        ax.set_ylabel(f'{selected_idyom_output}')
        ax.set_yscale('linear')

        if grid is True:
            ax.grid(True)
        else:
            ax.grid(False)
        return ax

    @staticmethod
    def one_ic_along_onsets(ax: matplotlib.axes.Axes,
                            melody_info: MelodyInfo,
                            chosen_ic: str,  # type (e.g., 'cpitch.information.content')
                            grid: bool,
                            color: str,
                            yticks):

        onset_in_beat = melody_info._get_onset_beat_nparray()
        chosen_ic_values = melody_info.get_idyom_output_nparray(idyom_output_key=chosen_ic)

        ax.plot(onset_in_beat, chosen_ic_values, marker='o', label=chosen_ic, color=color, markersize=5)

        xticks = np.arange(0, np.amax(onset_in_beat))
        yticks = yticks + 1

        ax.set_xticks(ticks=xticks, minor=True)
        # ax.set_xlabel('Time in quarter note')
        ax.set_yticks(ticks=yticks, minor=True)
        ax.legend()

        if grid is True:
            ax.grid(True)
        else:
            ax.grid(False)
        return ax

    @staticmethod
    def one_entropy_along_onsets(ax: matplotlib.axes.Axes,
                                 melody_info: MelodyInfo,
                                 chosen_entropy: str,  # type (e.g., 'cpitch.information.content')
                                 grid: bool,
                                 color: str,
                                 yticks):

        onset_in_beat = melody_info._get_onset_beat_nparray()
        chosen_entropy_values = melody_info.get_idyom_output_nparray(idyom_output_key=chosen_entropy)

        ax.plot(onset_in_beat, chosen_entropy_values, marker='o', label=chosen_entropy, color=color, markersize=5)

        xticks = np.arange(0, np.amax(onset_in_beat))
        yticks = yticks + 1

        ax.set_xticks(ticks=xticks, minor=True)
        # ax.set_xlabel('Time in quarter note')
        ax.set_yticks(ticks=yticks, minor=True)
        ax.legend()

        if grid is True:
            ax.grid(True)
        else:
            ax.grid(False)
        return ax

    @staticmethod
    def selected_ic_entropy_along_onsets(ax: matplotlib.axes.Axes,
                                         melody_info: MelodyInfo,
                                         ic_source: str,
                                         entropy_source: str,
                                         grid: bool,
                                         ic_color=None,
                                         entropy_color=None,
                                         ):

        valid_keywords_list = melody_info.get_idyom_output_keyword_list()
        valid_surprisal_source = [keyword for keyword in valid_keywords_list if 'information.content' in keyword]
        valid_entropy_source = [keyword for keyword in valid_keywords_list if 'entropy' in keyword]

        # check usr inputs:
        if ic_source in valid_surprisal_source:
            ic_source_values = melody_info.access_idyom_output_keywords([ic_source]).values.tolist()
            flatten_ic_values = [item for sublist in ic_source_values for item in sublist]
        else:
            raise ValueError(f'ic_source \'{ic_source}\' is invalid. '
                             f'Valid information content source are: {valid_surprisal_source}')

        if entropy_source in valid_entropy_source:
            entropy_source_values = melody_info.access_idyom_output_keywords([entropy_source]).values.tolist()
            flatten_entropy_values = [item for sublist in entropy_source_values for item in sublist]
        else:
            raise ValueError(f'entropy_source \'{entropy_source}\' is invalid. '
                             f'Valid entropy source are: {valid_entropy_source}')

        onset_in_beat = melody_info._get_onset_beat_nparray()

        ax.plot(onset_in_beat, flatten_ic_values, "-o", label=ic_source, color=ic_color)
        ax.plot(onset_in_beat, flatten_entropy_values, "-o", label=entropy_source, color=entropy_color)

        max_ic = np.amax(flatten_ic_values)
        max_entropy = np.amax(flatten_entropy_values)

        xticks = np.arange(0, np.amax(onset_in_beat))
        yticks = np.arange(0, max(max_ic, max_entropy) + 1)

        ax.set_xticks(ticks=xticks, minor=True)
        ax.set_xlabel('Time in quarter note')
        ax.set_yticks(ticks=yticks, minor=True)
        ax.legend()
        if grid is True:
            ax.grid(True)
        else:
            ax.grid(False)
        #    ax.xaxis.set_ticklabels([])  # hide xtick labels
        # ax.set_ylabel(f'{ic_source}')

        return ax

    @staticmethod
    def pianoroll(ax: matplotlib.axes.Axes,
                  melody_info: MelodyInfo):
        sustain_mask = melody_info._get_pianoroll_original()

        pitch_min = np.amin(melody_info.exp_pitch_element_list)
        pitch_max = np.amax(melody_info.exp_pitch_element_list)

        duration_in_ticks = sustain_mask.shape[1]
        onsets = melody_info.access_idyom_output_keywords(['onset']).to_numpy(dtype=int).reshape(-1)
        onsets_binary = np.zeros(duration_in_ticks)
        np.put(a=onsets_binary, ind=onsets, v=1)
        onsets_binary = onsets_binary.astype(bool)

        onsets_mask = onsets_binary * sustain_mask
        background_mask = np.bitwise_not(sustain_mask)

        colored_image = np.zeros(shape=sustain_mask.shape + (3,))
        colored_image[sustain_mask] = np.array([253, 231, 37])
        colored_image[onsets_mask] = np.array([59, 82, 139])
        colored_image[background_mask] = np.array([68, 1, 84])
        colored_image = colored_image.astype(int)

        ax.imshow(colored_image, origin='lower', aspect='auto',
                  extent=[0, duration_in_ticks / 24, pitch_min, pitch_max])
        # ax.axis('image')
        # ax.set_xlabel("Time in quarter note")
        # ax.xaxis.set_ticklabels([])  # hide xtick labels
        ax.set_ylabel('Pitch (MIDI number)', fontsize=10)
        return ax

    @staticmethod
    def pianoroll_pitch_distribution(ax: matplotlib.axes.Axes,
                                     melody_info: MelodyInfo):
        pitch_min = np.amin(melody_info.exp_pitch_element_list)
        pitch_max = np.amax(melody_info.exp_pitch_element_list)

        duration_in_ticks = melody_info._get_pianoroll_original().shape[1]
        pianoroll_distribution_array = melody_info._get_pianoroll_pitch_distribution()
        ax.imshow(pianoroll_distribution_array, origin='lower', aspect='auto',
                  extent=[0, duration_in_ticks / 24, pitch_min, pitch_max])

        # ax.title.set_text('Pitch Prediction')
        # ax.set_xlabel("Time in quarter note")
        ax.set_ylabel('Pitch (MIDI number)', fontsize=10)
        return ax

    @staticmethod
    def entropy(ax: matplotlib.axes.Axes,
                melody_info: MelodyInfo,
                color=None):

        entropy_values = np.int_(melody_info.access_properties(['entropy']))
        onset_values = np.int_(melody_info.access_properties(['onset']))
        ax.plot(onset_values, entropy_values, color=color)
        ax.margins(x=0.01)
        ax.margins(y=0)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_xlabel("Time in quarter note")
        #    ax.xaxis.set_ticklabels([])  # hide xtick labels
        ax.set_ylabel('Entropy', fontsize=10)

        return ax

    @staticmethod
    def surprisal(ax: matplotlib.axes.Axes,
                  melody_info: MelodyInfo,
                  color=None):

        surprisal_values = np.int_(melody_info.access_properties(['information.content']))
        onset_values = np.int_(melody_info.access_properties(['onset']))
        ax.plot(onset_values, surprisal_values, color=color)
        ax.margins(x=0.01)
        ax.margins(y=0)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_ylabel('Information Content (Surprisal) \n -log(P)', fontsize=10)
        return ax

    @staticmethod
    def surprisal_continuous(ax: matplotlib.axes.Axes,
                             melody_info: MelodyInfo,
                             color=None):

        # spike at onset
        surprisal_array = melody_info._get_surprisal_array()
        onset_time_vector = melody_info._get_onset_time_vector() / 24
        ax.plot(onset_time_vector, surprisal_array, color=color)
        ax.margins(x=0.01)
        ax.margins(y=0)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        # ax.set_xlabel('Time in quarter note')
        #    ax.xaxis.set_ticklabels([])  # hide xtick labels
        ax.set_ylabel('Information Content (Surprisal) \n -log(P)', fontsize=10)
        return ax


class Auxiliary:

    @staticmethod
    def batch_melodies_plots(plot_method_func,
                             fig_format: str,
                             dpi: float,
                             plot_type_folder_name: str,
                             experiment_folder_path: str,
                             melody_names: List[str] = None,
                             starting_index: int = None,
                             ending_index: int = None,
                             savefig: bool = True,
                             ):

        experiment_info = ExperimentInfo(experiment_folder_path=experiment_folder_path)
        all_melody_names = list(experiment_info.melodies_dict.keys())
        print(plot_type_folder_name)
        saved_msg = str('Plots saved in ' + experiment_folder_path + 'plots/' + str(plot_type_folder_name) + '/')

        def _common_batch_actions():
            melody_info = experiment_info.melodies_dict[melody]
            melody_name_pprint = melody_info._get_melody_name_pprint()
            fig = plot_method_func(melody_info)
            if savefig is True:
                Auxiliary.save_one_fig(plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_name_pprint=melody_name_pprint,
                                       fig=fig,
                                       fig_format=fig_format,
                                       dpi=dpi)
                print(saved_msg)

        if melody_names:
            for index, melody in enumerate(melody_names):
                _common_batch_actions()

        elif starting_index or ending_index:
            for index, melody in enumerate(all_melody_names[starting_index:ending_index]):
                _common_batch_actions()

        else:
            for index, melody in enumerate(all_melody_names):
                _common_batch_actions()

    @staticmethod
    def save_one_fig(plot_type_folder_name: str,
                     experiment_folder_path: str,
                     melody_name_pprint: str,
                     fig,
                     fig_format: str,
                     dpi: float):
        output_path = experiment_folder_path + 'plots/' + plot_type_folder_name + '/'
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        fig.savefig(fname=output_path + str(melody_name_pprint) + '.' + fig_format, format=fig_format, dpi=dpi)


class BasicPlot:
    """ BasicPlot for selected IDyOM outputs in an experiment. """

    @staticmethod
    def pianoroll_pitch_prediction_groundtruth(experiment_folder_path: str,
                                               melody_names: List[str] = None,
                                               starting_index: int = None,
                                               ending_index: int = None,
                                               savefig: bool = True,
                                               showfig: bool = False,
                                               fig_format: str = 'png',
                                               dpi: float = 400,
                                               figsize: tuple = (10, 10),
                                               nrows: int = 2,
                                               ncols: int = 1, ):

        """
        Generate a pair of figures (the predicted pitch distribution and the ground truth) side by side.
        If users intend to plot figures for specific songs, they can do so by specifying either the melody names,
        or the starting/ending index in the melody list.

        :param experiment_folder_path: The path to your experiment folder.
        :param melody_names: (optional), if not supplied by the users, default is all test melodies in the experiment.
        :param starting_index: (optional), the index of the melody in the melody list that you want to start plotting.
        :param ending_index: (optional), the index of the melody in the melody list that you want to stop plotting.
        :param savefig: (optional)
        :param showfig: (optional)
        :param fig_format: (optional), default = 'png
        :param dpi: (optional), default = 400
        :param figsize: (optional), default is (10,5)
        :param ncols: (optional), the number of columns of the figure. By default, ncols = 2, nrows = 1, figures are shown side-by-side.
        :param nrows: (optional), the number of columns of the figure. By default, ncols = 2, nrows = 1, figures are shown side-by-side.

        """

        plot_type_folder_name = 'pianoroll_pitch_prediction_groundtruth'

        def _pianoroll_pitch_prediction_groundtruth(melody_info: MelodyInfo,
                                                    show_single_fig: bool = showfig) -> plt.Figure:

            melody_name_pprint = melody_info._get_melody_name_pprint()

            fig, (ax_distribution, ax_groundtruth) = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, dpi=dpi)
            fig.suptitle('IDyOM Pitch prediction vs Ground truth \n\n Melody name: ' + melody_name_pprint)

            BasicAxsGeneration.pianoroll_pitch_distribution(ax_distribution, melody_info=melody_info)
            BasicAxsGeneration.pianoroll(ax_groundtruth, melody_info=melody_info)
            # ax_groundtruth.title.set_text('Ground Truth')

            plt.xlabel("Time in quarter note")
            plt.tight_layout()

            if show_single_fig is True:
                plt.show()
            else:
                pass

            return fig

        Auxiliary.batch_melodies_plots(plot_method_func=_pianoroll_pitch_prediction_groundtruth,
                                       plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_names=melody_names,
                                       starting_index=starting_index,
                                       ending_index=ending_index,
                                       savefig=savefig,
                                       fig_format=fig_format,
                                       dpi=dpi
                                       )

    @staticmethod
    def pianoroll_groundtruth_overall_surprisal(experiment_folder_path: str,
                                                melody_names: List[str] = None,
                                                starting_index: int = None,
                                                ending_index: int = None,
                                                savefig: bool = True,
                                                showfig: bool = False,
                                                fig_format: str = 'png',
                                                dpi: float = 400,
                                                figsize: tuple = (10, 6)):
        """
        Generate a pair of figures: ground truth piano roll on the top and the surprisal line plot on the bottom.

        :param experiment_folder_path: the path to your experiment folder.
        :param melody_names: (optional), if not supplied by the users, default is all test melodies in the experiment.
        :param starting_index: (optional),the index of the melody in the melody list that you want to start plotting.
        :param ending_index: (optional), the index of the melody in the melody list that you want to stop plotting.
        :param savefig: (optional)
        :param showfig: (optional)
        :param fig_format: (optional), default = 'png'
        :param dpi: (optional), default = 400
        :param figsize: (optional), default is (10,5)

        """

        plot_type_folder_name = 'pianoroll_groundtruth_surprisal'

        def _pianoroll_groundtruth_surprisal(melody_info: MelodyInfo,
                                             show_single_fig: bool = showfig) -> plt.Figure:

            melody_name_pprint = melody_info._get_melody_name_pprint()

            fig, (ax_pianoroll, ax_surprisal) = plt.subplots(2, 1, figsize=figsize, dpi=dpi, sharex='col')
            fig.suptitle('IDyOM Information Content (Surprisal) \n\n Melody: ' + melody_name_pprint)

            BasicAxsGeneration.pianoroll(ax_pianoroll, melody_info=melody_info)
            BasicAxsGeneration.surprisal_continuous(ax_surprisal, melody_info=melody_info)

            plt.xlabel("Time in quarter note")
            plt.tight_layout()

            if show_single_fig is True:
                plt.show()
            else:
                pass
            return fig

        Auxiliary.batch_melodies_plots(plot_method_func=_pianoroll_groundtruth_surprisal,
                                       plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_names=melody_names,
                                       starting_index=starting_index,
                                       ending_index=ending_index,
                                       savefig=savefig,
                                       fig_format=fig_format,
                                       dpi=dpi)

    @staticmethod
    def simple_plot(selected_idyom_output: str,
                    experiment_folder_path: str,
                    melody_names: List[str] = None,
                    starting_index: int = None,
                    ending_index: int = None,
                    savefig: bool = True,
                    showfig: bool = False,
                    fig_format: str = 'png',
                    dpi: float = 400,
                    figsize: tuple = (10, 5),
                    grid: bool = True,
                    ggplot: bool = True):
        """
        Generate a simple line plot with time (in quarter note) on the x-axis, and selected IDyOM output on the y-axis.

        :param selected_idyom_output: the keyword of the IDyOM output you want to plot.
        :param experiment_folder_path: the path to your experiment folder.
        :param melody_names: if not supplied by the users, default is all test melodies in the experiment.
        :param starting_index: the index of the melody in the melody list that you want to start plotting.
        :param ending_index: the index of the melody in the melody list that you want to stop plotting.
        :param savefig:
        :param showfig:
        :param fig_format: default = 'png
        :param dpi: optional, default = 400
        :param figsize: optional, default is (10,5)
        :param grid: whether to show grid or not.
        :param ggplot: whether to use ggplot or not.

        """
        plot_type_folder_name = 'simple_plot_' + selected_idyom_output

        def _generic_property_along_time(melody_info: MelodyInfo,
                                         show_single_fig: bool = showfig,
                                         grid: bool = grid,
                                         ggplot: bool = ggplot) -> plt.Figure:

            if ggplot is True:
                plt.style.use('ggplot')
            else:
                pass

            melody_name_pprint = melody_info._get_melody_name_pprint()
            fig, (ax_generic_property) = plt.subplots(nrows=1, ncols=1, figsize=figsize, dpi=dpi, sharex='col')

            fig.suptitle('Selected IDyOM outputs: ' + selected_idyom_output + '\n\n Melody: ' + melody_name_pprint,
                         fontsize=15)
            BasicAxsGeneration.generic_idyom_output_along_time(ax_generic_property,
                                                               melody_info=melody_info,
                                                               selected_idyom_output=selected_idyom_output,
                                                               grid=grid)
            plt.tight_layout()
            if show_single_fig is True:
                plt.show()
            else:
                pass
            return fig

        # Plot batch according to user inputs:
        Auxiliary.batch_melodies_plots(plot_method_func=_generic_property_along_time,
                                       plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_names=melody_names,
                                       starting_index=starting_index,
                                       ending_index=ending_index,
                                       savefig=savefig,
                                       fig_format=fig_format,
                                       dpi=dpi)

    @staticmethod
    def selected_surprisal_entropy(experiment_folder_path: str,
                                   ic_source: str,
                                   entropy_source: str,
                                   melody_names: List[str] = None,
                                   starting_index: int = None,
                                   ending_index: int = None,
                                   savefig: bool = True,
                                   showfig: bool = False,
                                   fig_format: str = 'png',
                                   dpi: float = 400,
                                   figsize: tuple = (10, 6),
                                   grid: bool = True,
                                   ggplot: bool = True):
        """
        Generate a figure that shows the selected entropy and information content.

        :param experiment_folder_path: the path to your experiment folder.
        :param entropy_source: the source of entropy.
        :param ic_source: the source of information content (surprisal)
        :param melody_names: if not supplied by the users, default is all test melodies in the experiment.
        :param starting_index: the index of the melody in the melody list that you want to start plotting.
        :param ending_index: the index of the melody in the melody list that you want to stop plotting.
        :param savefig:
        :param showfig:
        :param fig_format: default = 'png
        :param dpi: default = 400
        :param figsize: default is (10,5)
        :param grid: whether to show grid or not.
        :param ggplot: whether to use ggplot or not.

        """

        plot_type_folder_name = 'selected_surprisal_entropy'

        def _selected_surprisal_entropy(melody_info: MelodyInfo,
                                        ic_source: str = ic_source,
                                        entropy_source: str = entropy_source,
                                        show_single_fig: bool = showfig,
                                        grid: bool = grid,
                                        ggplot: bool = ggplot) -> plt.Figure:

            if ggplot is True:
                plt.style.use('ggplot')
            else:
                pass
            melody_name_pprint = melody_info._get_melody_name_pprint()
            fig, ax = plt.subplots(figsize=figsize)
            fig.suptitle('IDyOM Information Content (Surprisal) and Entropy \n\n Melody: ' + melody_name_pprint)

            BasicAxsGeneration.selected_ic_entropy_along_onsets(ax=ax,
                                                                melody_info=melody_info,
                                                                ic_source=ic_source,
                                                                entropy_source=entropy_source,
                                                                ic_color='C0',
                                                                entropy_color='C1',
                                                                grid=grid)

            plt.tight_layout()

            if show_single_fig is True:
                plt.show()
            else:
                pass
            return fig

        Auxiliary.batch_melodies_plots(plot_method_func=_selected_surprisal_entropy,
                                       plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_names=melody_names,
                                       starting_index=starting_index,
                                       ending_index=ending_index,
                                       savefig=savefig,
                                       fig_format=fig_format,
                                       dpi=dpi)

    @staticmethod
    def all_surprisal(experiment_folder_path: str,
                      melody_names: List[str] = None,
                      starting_index: int = None,
                      ending_index: int = None,
                      savefig: bool = True,
                      showfig: bool = False,
                      fig_format: str = 'png',
                      dpi: float = 400,
                      figsize: tuple = (10, 8),
                      grid: bool = True,
                      ggplot: bool = True):
        """
        Generate subplots of all available surprisal outputs.

        :param experiment_folder_path: the path to your experiment folder.
        :param melody_names: if not supplied by the users, default is all test melodies in the experiment.
        :param starting_index: the index of the melody in the melody list that you want to start plotting.
        :param ending_index: the index of the melody in the melody list that you want to stop plotting.
        :param savefig:
        :param showfig:
        :param fig_format: default = 'png
        :param dpi: default = 400
        :param figsize: default is (10,5)
        :param grid: whether to show grid or not.
        :param ggplot: whether to use ggplot or not.

        """

        plot_type_folder_name = 'surprisals_plots'

        def _all_surprisal_plots(melody_info: MelodyInfo,
                                 grid: bool = grid,
                                 show_single_fig: bool = showfig,
                                 figsize: tuple = figsize,
                                 dpi: float = dpi,
                                 ggplot: bool = ggplot) -> plt.Figure:

            if ggplot is True:
                plt.style.use('ggplot')
            else:
                pass

            melody_name_pprint = melody_info._get_melody_name_pprint()

            valid_keywords_list = melody_info.get_idyom_output_keyword_list()
            all_surprisal_sources = [keyword for keyword in valid_keywords_list if 'information.content' in keyword]

            num_of_surprisal_subplots = len(all_surprisal_sources)
            # to get the same y-range for all plots (0, max_surprisal across all surprisals)
            max_surprisals = []
            for idx, surprisal_source in enumerate(all_surprisal_sources):
                surprisals = np.amax(melody_info.get_idyom_output_nparray(surprisal_source))
                max_surprisals.append(surprisals)
            yticks = np.arange(0, max(max_surprisals))

            # plot a number of surprisal plots stacked one on top of each other, shared x, y
            # add every single subplot to the figure with a for loop
            fig, axs = plt.subplots(num_of_surprisal_subplots, figsize=figsize, dpi=dpi, sharex='col', sharey='row')
            fig.suptitle('IDyOM Information Content (Surprisal) \n\n Melody: ' + melody_name_pprint)

            for idx, surprisal_type in enumerate(all_surprisal_sources):
                BasicAxsGeneration.one_ic_along_onsets(ax=axs[idx],
                                                       melody_info=melody_info,
                                                       chosen_ic=surprisal_type,
                                                       grid=grid,
                                                       color=str('C' + str(idx)),
                                                       yticks=yticks)

            plt.xlabel("Time in quarter note")
            plt.tight_layout()
            plt.style.use('ggplot')

            if show_single_fig is True:
                plt.show()
            else:
                pass
            return fig

        Auxiliary.batch_melodies_plots(plot_method_func=_all_surprisal_plots,
                                       plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_names=melody_names,
                                       starting_index=starting_index,
                                       ending_index=ending_index,
                                       savefig=savefig,
                                       fig_format=fig_format,
                                       dpi=dpi)

    @staticmethod
    def all_entropy(experiment_folder_path: str,
                    melody_names: List[str] = None,
                    starting_index: int = None,
                    ending_index: int = None,
                    savefig: bool = True,
                    showfig: bool = False,
                    fig_format: str = 'png',
                    dpi: float = 400,
                    figsize: tuple = (10, 8),
                    grid: bool = True,
                    ggplot: bool = True):
        """
        Generate subplots that show all available entropy outputs.

        :param experiment_folder_path: the path to your experiment folder.
        :param melody_names: if not supplied by the users, default is all test melodies in the experiment.
        :param starting_index: the index of the melody in the melody list that you want to start plotting.
        :param ending_index: the index of the melody in the melody list that you want to stop plotting.
        :param savefig:
        :param showfig:
        :param fig_format: default = 'png
        :param dpi: default = 400
        :param figsize: default is (10,5)
        :param grid: whether to show grid or not.
        :param ggplot: whether to use ggplot or not.

        """

        plot_type_folder_name = 'entropy_plots'

        def _all_entropy_plots(melody_info: MelodyInfo,
                               grid: bool = grid,
                               show_single_fig: bool = showfig,
                               figsize: tuple = figsize,
                               dpi: float = dpi,
                               ggplot: bool = ggplot) -> plt.Figure:

            if ggplot is True:
                plt.style.use('ggplot')
            else:
                pass

            melody_name_pprint = melody_info._get_melody_name_pprint()

            valid_keywords_list = melody_info.get_idyom_output_keyword_list()
            all_entropy_sources = [keyword for keyword in valid_keywords_list if 'entropy' in keyword]

            num_of_surprisal_subplots = len(all_entropy_sources)
            # to get the same y-range for all plots (0, max_surprisal across all surprisals)
            max_entropys = []
            for idx, entropy_source in enumerate(all_entropy_sources):
                entropys = np.amax(melody_info.get_idyom_output_nparray(entropy_source))
                max_entropys.append(entropys)
            yticks = np.arange(0, max(max_entropys))

            # plot a number of entropy plots stacked one on top of each other, shared x, y
            # add every single subplot to the figure with a for loop
            fig, axs = plt.subplots(num_of_surprisal_subplots, figsize=figsize, dpi=dpi, sharex='col', sharey='row')
            fig.suptitle('IDyOM Entropy \n\n Melody: ' + melody_name_pprint)

            for idx, entropy_type in enumerate(all_entropy_sources):
                BasicAxsGeneration.one_entropy_along_onsets(ax=axs[idx],
                                                            melody_info=melody_info,
                                                            chosen_entropy=entropy_type,
                                                            grid=grid,
                                                            color=str('C' + str(idx)),
                                                            yticks=yticks)

            plt.xlabel("Time in quarter note")
            plt.tight_layout()
            plt.style.use('ggplot')

            if show_single_fig is True:
                plt.show()
            else:
                pass
            return fig

        Auxiliary.batch_melodies_plots(plot_method_func=_all_entropy_plots,
                                       plot_type_folder_name=plot_type_folder_name,
                                       experiment_folder_path=experiment_folder_path,
                                       melody_names=melody_names,
                                       starting_index=starting_index,
                                       ending_index=ending_index,
                                       savefig=savefig,
                                       fig_format=fig_format,
                                       dpi=dpi)



import typing
from dataclasses import dataclass
from glob import glob

import numpy as np
import pandas as pd


def to_float(f):
    try:
        return float(f)
    except ValueError:
        return f


def get_dictionary(file: str) -> dict:
    """
    Read the file line by line, split each line into n fields, then create the dictionary:
    :param: file
    :return: dict
    """

    dict = {}
    f = open(file, "r")
    lines = f.readlines()
    keys = lines[0].split()
    for i in range(1, len(lines)):
        fields = lines[i].split()

        if fields[1] not in dict:
            dict[fields[1]] = {}  # create dict for each melody

        k = 0
        for key in keys:
            if key not in dict[fields[1]]:
                dict[fields[1]][key] = []
            dict[fields[1]][key].append(to_float(fields[k]))
            k += 1
    return dict


def get_all_song_dict(dat_file_path: str) -> dict:
    all_song_dict = get_dictionary(dat_file_path)
    return all_song_dict


def get_song_dict_of_interest(all_song_dict, melody_id):
    return list(all_song_dict.values())[melody_id]


def getDataFrame(file: str) -> pd.DataFrame:
    df = pd.read_table(file, delim_whitespace=True)
    return df


class MelodyInfo(pd.DataFrame):
    """
    A melody object (pd.DataFrame) that contains all data in a single melody inherit from the parent Experiment.

    """

    _metadata = ['exp_pitch_element_list', 'parent_experiment']

    def __init__(self, exp_pitch_element_list, parent_experiment, *args, **kw):
        super().__init__(*args, **kw)
        self.exp_pitch_element_list = exp_pitch_element_list
        self.parent_experiment = parent_experiment
        self.melody_name_pp = self._get_melody_name_pprint()

    def access_idyom_output_keywords(self, output_keywords: typing.List[str]):
        """
        Access certain idyom output(s) via its (their) keyword(s).

        :param output_keywords: A list of IDyOM output keywords (e.g., ['cpitch.information.content', 'onset', 'entropy'])
        :return: pd.DataFrame
        """

        if isinstance(output_keywords, list):
            pass
        else:
            raise TypeError(f'Argument \'output_keywords\' should be a list of strings, not {type(output_keywords)}')

        def check_condition(keyword) -> bool:
            if keyword in self.get_idyom_output_keyword_list():
                return True
            else:
                raise KeyError(
                    f'Incorrect keyword: \'{keyword}\'. Please check your spelling. Available IDyOM output keywords for this melody are: {self.get_idyom_output_keyword_list()}')

        for idx, keyword in enumerate(output_keywords):
            if check_condition(keyword):
                return self[output_keywords]

    def get_idyom_output_nparray(self, idyom_output_key: str):
        """
        Get the IDyOM output via its key as a np.array

        :param idyom_output_key: list of str
        :return: np.array of output values
        """

        output_values = (self.access_idyom_output_keywords([idyom_output_key]).values.tolist())
        output_values_array = [item for sublist in output_values for item in sublist]
        return output_values_array

    def get_idyom_output_keyword_list(self) -> list:
        """
        Get a list of available IDyOM output keyword for this melody.

        :return: list
        """

        idyom_output_keyword_list = self.keys().to_list()
        return idyom_output_keyword_list

    def compute_properties_means(self, idyom_outputs: typing.List[str]):
        """
        Compute the mean values of the idyom outputs.

        :param idyom_outputs: list of strings
        :return: DataFrame
        """

        cropped_df = self[idyom_outputs]
        return cropped_df.mean(axis=0)

    def _get_onset_time_in_seconds(self):
        onset_values = np.int_(self.access_idyom_output_keywords(['onset']))
        base_onset_values = onset_values / 24  # idyom uses basic time units, quarter note =24
        tempo = np.int_(self.access_idyom_output_keywords(['tempo']))
        bpm = (60 * 1000000) / tempo
        onset_time_in_sec = (bpm / 60) * base_onset_values
        return onset_time_in_sec

    def _get_onset_beat_nparray(self):
        onset_values = np.int_(self.access_idyom_output_keywords(['onset']).values.tolist())
        onset_in_beat = onset_values / 24
        onset_in_beat_array = [item for sublist in onset_in_beat for item in sublist]
        return onset_in_beat_array

    def _get_melody_name_pprint(self) -> str:
        melody_name_pprint = str(self.access_idyom_output_keywords(['melody.name']).to_numpy()[0][0]).replace('"', '')
        return melody_name_pprint

    def _get_pianoroll_pitch_distribution(self):
        pitch_range = (np.amin(self.exp_pitch_element_list), np.amax(self.exp_pitch_element_list))
        pitch_distribution = []
        melody_length = len(self.access_idyom_output_keywords(['cpitch.probability']))
        for i in range(*pitch_range):
            key_to_look = 'cpitch.' + str(i)
            if key_to_look not in self.keys():
                one_pitch_prob_across_time = [0] * melody_length
            else:
                one_pitch_prob_across_time = self[key_to_look]
            pitch_distribution.append(one_pitch_prob_across_time)
        pitch_distribution = np.array(pitch_distribution)
        durations = self.access_idyom_output_keywords((['dur'])).to_numpy(dtype=int).reshape(-1)
        pitch_distribution = np.repeat(pitch_distribution, repeats=durations, axis=1)
        return pitch_distribution

    def _get_pianoroll_original(self):
        pitch_range = (np.amin(self.exp_pitch_element_list), np.amax(self.exp_pitch_element_list))
        piano_roll = np.arange(*pitch_range)
        piano_roll = (piano_roll == self.access_idyom_output_keywords(['cpitch']).to_numpy()).T
        durations = self.access_idyom_output_keywords((['dur'])).to_numpy(dtype=int).reshape(-1)
        piano_roll = np.repeat(piano_roll, repeats=durations, axis=1)
        return piano_roll

    def _get_onset_time_vector(self):
        onset_seq = self.access_idyom_output_keywords(['onset'])
        onset_seq_int = np.int_(onset_seq)
        onset_time_vector = np.arange(0, onset_seq_int[-1] + 1)
        return onset_time_vector

    def _get_surprisal_array(self):
        # ic_seq = surprisal seq
        onset_seq_int = np.int_(self.access_idyom_output_keywords(['onset']))
        onset_time_vector = self._get_onset_time_vector()
        extended_ic_seq = np.zeros(len(onset_time_vector))

        ic_seq = self.access_idyom_output_keywords(['information.content'])
        np.put(extended_ic_seq, onset_seq_int, ic_seq)
        return extended_ic_seq


@dataclass
class ExperimentInfo:
    """
    An experiment object that contains all data in a single experiment.

    :param experiment_folder_path: The path to experiment log folder which you want to access.
    """

    experiment_folder_path: str

    def __post_init__(self):
        self.dat_file_path = sorted(glob(self.experiment_folder_path + 'experiment_output_data_folder/*'))[0]
        self.df = pd.read_table(self.dat_file_path, delim_whitespace=True)
        self.exp_pitch_element_list = self._get_datasetwise_cpitch_elements()
        self.melodies_dict = self.melody_dictionary()

    def melody_dictionary(self) -> typing.Dict[str, MelodyInfo]:
        """
        Get a dictionary of all melodies in the experiment with melody name as the key and all melody info as the value.

        :return: a typed dictionary (melody_name, MelodyInfo)
        """

        return_dict = {}
        my_dict = get_dictionary(self.dat_file_path)
        for key, value in my_dict.items():
            melody_info = MelodyInfo(data=value, parent_experiment=self,
                                     exp_pitch_element_list=self._get_datasetwise_cpitch_elements())
            melody_name = melody_info['melody.name'][0]
            return_dict[melody_name] = melody_info
        return return_dict

    def access_melodies(self, starting_index=None, ending_index=None,
                        melody_names=None):
        """
        Access specific melodies by index or melody names.
        If all arguments are None, then the default is to access all melodies in the Experiment class

        :param starting_index: int
        :param ending_index: int
        :param melody_names: list of str
        :return: a list of MelodyInfo class objects (selected melodies)
        """

        if melody_names is not None:
            selected_melodies = list(map(self.melodies_dict.get, melody_names))
        else:
            selected_melodies = list(self.melodies_dict.values())[starting_index:ending_index]

        return selected_melodies

    def _get_datasetwise_cpitch_elements(self):
        """
        Get the list of cpitch (full cpitch distribution elements used in IDyOM)

        :return:  a list of int
        """
        # find the cpitches in idyom output keys such as 'cpitch.

        df_all_keys = self.df.keys().to_list()
        cpitch_keys = [keyword for keyword in df_all_keys if 'cpitch' in keyword]
        cpitch_num_keys = [item for item in cpitch_keys if any([char.isdigit() for char in item])]
        pitch_element_list = [item.replace('cpitch.', '') for item in cpitch_num_keys]
        pitch_element_list = np.int_(pitch_element_list)

        return pitch_element_list



import os
from dataclasses import dataclass
from glob import glob
from typing import List

import numpy as np
import pandas as pd
import scipy.io

from py2lispIDyOM.extract import ExperimentInfo


@dataclass
class Export:
    """Export selected IDyOM model outputs to other formats.

    :param experiment_folder_path: The path to which you saved all the result data/plots
    :param idyom_output_keywords: A list of IDyOM output keywords you want to export, defaults to None
    :param melody_names: A list of melodies of which IDyOM outputs that you want to export
    """

    experiment_folder_path: str
    idyom_output_keywords: List = None
    melody_names: List = None

    def __post_init__(self):
        self.dat_file_path = glob(self.experiment_folder_path + 'experiment_output_data_folder/*.dat')[0]
        self.experiment_info = ExperimentInfo(experiment_folder_path=self.experiment_folder_path)
        self.melodies_info_dict = self.experiment_info.melodies_dict

    def _generate_export_folder(self, export_folder_name):
        """To generate a folder to store the idyom outputs in other formats (e.g., .mat, .csv)"""
        idyom_output_export_folder_path = self.experiment_folder_path + export_folder_name + '/'
        if not os.path.exists(idyom_output_export_folder_path):
            os.makedirs(idyom_output_export_folder_path)
        return idyom_output_export_folder_path

    def _get_valid_idyom_output_keys(self, melody):
        """To get a list of valid idyom output keys for a melody."""
        valid_idyom_output_keys = self.melodies_info_dict[melody].get_idyom_output_keyword_list()
        return valid_idyom_output_keys

    def _get_idyom_output_for_single_melody(self, melody, idyom_key):
        """To get the IDyOM output value array for a single melody, according to the IDyOM output keys."""
        output_value_array = self.melodies_info_dict[melody].access_idyom_output_keywords(
            [idyom_key]).values  # this is np.array
        output_value_array = output_value_array.flatten()
        return output_value_array

    def _get_single_melody_output_values_df(self, melody):
        """This function returns a DataFrame of all IDyOM output values of one melody."""
        idyom_output_values_df = self.melodies_info_dict[melody]
        return idyom_output_values_df

    def _get_output_values_in_selected_melodies(self, idyom_key, selected_songs):
        """This function returns a np.array of idyom output values for selected an idyom key for each selected songs."""
        output_values_data_in_songs = []
        for index, melody in enumerate(selected_songs):
            valid_keys = self._get_valid_idyom_output_keys(melody=melody)
            if idyom_key in valid_keys:
                idyom_data = self._get_idyom_output_for_single_melody(melody=melody, idyom_key=idyom_key)
                output_values_data_in_songs.append(idyom_data)
            else:
                raise KeyError(
                    f'IDyOM output keyword \'{idyom_key}\' is invalid. Valid IDyOM output keys are: {valid_keys}')
        output_values_data_in_songs = np.array(output_values_data_in_songs, dtype=object)
        return output_values_data_in_songs

    def _get_output_values_df_in_selected_melodies(self, selected_songs):
        """
        This function iterate through the func '_get_single_melody_output_values_df' to get
        a DataFrame of idyom output values for all idyom keys for each selected songs.
         """
        for index, melody in selected_songs:
            self._get_single_melody_output_values_df(melody=melody)

    def _export_by_keyword_2mat(self, keywords_list, selected_songs, output_path):
        """Export the idyom output data to .mat files according to the keyword list."""

        # Type check =====================:
        if isinstance(keywords_list, list):
            pass
        else:
            raise TypeError(f'Argument \'keywords_list\' should be a list of strings, not {type(keywords_list)}')

        for i, keyword in enumerate(keywords_list):  # data_to_export is a list of list
            keyword_output_data_in_songs = self._get_output_values_in_selected_melodies(idyom_key=keyword,
                                                                                        selected_songs=selected_songs)
            keyword_name_pp = keyword.replace('.', '_')  # account for names like "information.content"
            keyword_name_pp = keyword_name_pp.replace('-', '')
            scipy.io.savemat(output_path + keyword_name_pp + '.mat',
                             mdict={keyword_name_pp: np.array(keyword_output_data_in_songs)})
        print('Exported data to ' + output_path)

    def _export_values_of_keywords_by_melody_2mat(self, keywords_list, melody, output_path):
        """Exports the IDyOM output values according to the keywords_list of one song to a mat file."""
        # Type check =====================:
        if isinstance(keywords_list, list):
            pass
        else:
            raise TypeError(f'Argument \'keywords_list\' should be a list of strings, not {type(keywords_list)}')
        melody_name_pp = melody.replace('"', '')
        for i, keyword in enumerate(keywords_list):
            idyom_output_data_in_song = self._get_idyom_output_for_single_melody(melody=melody, idyom_key=keyword)
            idyom_keyword_pp = keyword.replace('.', '_')  # account for names like "information.content"
            full_outfile_name = melody_name_pp + '_' + idyom_keyword_pp
            full_outfile_name = full_outfile_name.replace('-', '')
            scipy.io.savemat(output_path + full_outfile_name + '.mat',
                             mdict={idyom_keyword_pp: np.array(idyom_output_data_in_song)})
        print('Exported data to ' + output_path)

    def _export_by_song_2csv(self, melody_name, single_song_df_data: pd.DataFrame, output_path):
        """Exports the all the IDyOM output data (df) for a single melody to csv"""
        melody_name = melody_name.replace('"', '')
        csv_file_path = output_path + melody_name + '.csv'
        single_song_df_data.to_csv(path_or_buf=csv_file_path, index=False, header=True)

    def export2mat(self):
        """
        This function exports the IDyOM output data to mat files.
        By default, it will export the outputs according to the preset keywords specified in the idyom_output_keywords for all melodies.
        Users can also specify specific melody by passing the melody names to the melody_name param.
        """
        # Check idyom_output_keyword is not None:
        if self.idyom_output_keywords is None:
            raise ValueError(f'The argument \'idyom_output_keywords\' is empty. '
                             f'Please provide a list of IDyOM output keywords of which you want to export to mat files.')
        else:
            pass
        export_folder_path = self._generate_export_folder(export_folder_name='outputs_in_mat')
        keywords = self.idyom_output_keywords

        if keywords:
            if self.melody_names:
                for index, melody in enumerate(self.melody_names):
                    self._export_values_of_keywords_by_melody_2mat(keywords_list=keywords, melody=melody,
                                                                   output_path=export_folder_path)

            else:
                melody_names = list(self.melodies_info_dict.keys())
                self._export_by_keyword_2mat(keywords_list=keywords, selected_songs=melody_names,
                                             output_path=export_folder_path)

    def export2csv(self):
        """
        This function exports the IDyOM output data to a csv files.
        By default, this will export all properties of the chosen melodies, where each melody is in one csv file.
        """
        # Check idyom_output_keyword is None:
        if self.idyom_output_keywords is not None:
            raise ValueError(f'The argument \'idyom_output_keywords\' should be empty. '
                             f'The current version of py2lispIDyoM only supports exporting all IDyoM outputs of selected or all melodies.')

        else:
            pass

        export_folder_path = self._generate_export_folder(export_folder_name='outputs_in_csv')

        if self.melody_names:
            for index, melody in enumerate(self.melody_names):
                single_song_df_data = self._get_single_melody_output_values_df(melody=melody)
                self._export_by_song_2csv(melody_name=melody, single_song_df_data=single_song_df_data,
                                          output_path=export_folder_path)

        else:
            melody_names = list(self.melodies_info_dict.keys())
            for index, melody in enumerate(melody_names):
                single_song_df_data = self._get_single_melody_output_values_df(melody=melody)
                self._export_by_song_2csv(melody_name=melody, single_song_df_data=single_song_df_data,
                                          output_path=export_folder_path)

        print('Exported data to ' + export_folder_path)



import setuptools

long_description = 'py2lispIDyOM is a Python package for the information dynamics of music (IDyOM) model.'

with open("requirements.txt", "r") as fh:
    install_requires = fh.read().splitlines()

setuptools.setup(
    name="py2lispIDyOM",
    version="1.0.0",
    author="Xinyi Guan",
    author_email="xinyi.guan@nyu.edu",
    description="A Python package for IDyOM model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xinyiguan/py2lispIDyOM",
    packages=setuptools.find_packages(exclude=['tests']),
    install_requires=install_requires,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



from __future__ import annotations

import datetime
import json
import os
import shutil
from abc import ABC, abstractmethod
from dataclasses import dataclass
from dataclasses import field
from glob import glob
from typing import Literal, List, Union, Tuple, Iterable, get_type_hints

from natsort import natsorted


def check_recursive_typings(obj, type_expected: type) -> bool:
    type_got = type(obj)

    if hasattr(type_expected, '__origin__'):
        type_origin = type_expected.__origin__
        if type_origin == Literal:  # type_expected is typing.Literal
            type_correct = obj in type_expected.__args__
        elif type_origin in [list, tuple]:
            arg_type = type_expected.__args__[0]
            type_correct = all([
                check_recursive_typings(obj=_, type_expected=arg_type) for _ in obj
            ])
        elif type_origin == Union:
            possible_types = type_expected.__args__
            type_correct = any([
                check_recursive_typings(obj=obj, type_expected=possible_type) for possible_type in possible_types
            ])
        else:
            raise TypeError()
    else:
        type_correct = type_got is type_expected
    return type_correct


def get_timestamp():
    today_date = datetime.date.today()
    now_time = datetime.datetime.now()
    moment = today_date.strftime('%m%d%y') + now_time.strftime('%H%M%S')
    return moment


@dataclass
class LispConvertable(ABC):
    @abstractmethod
    def to_lisp_command(self):
        pass

    def __repr__(self):
        return json.dumps(self, default=lambda x: x.__dict__, indent=2)

    def recursive_set_attr(self, key, value):
        # print(f'set {key=} {value=}')
        children_configuration = filter(lambda x: isinstance(x, Configuration), self.__dict__.values())
        children_parameter = filter(lambda x: isinstance(x, Parameters), self.__dict__.values())
        children = list(children_configuration) + list(children_parameter)

        if hasattr(self, key):
            # print(f'{self} has {key=}')
            type_hint_dict = get_type_hints(self, globalns=globals())
            if key not in type_hint_dict:
                print(f'type hint for {key} not defined')
            else:
                type_expected = type_hint_dict[key]
                type_correct = check_recursive_typings(obj=value, type_expected=type_expected)

                if type_correct:

                    self.__dict__[key] = value
                    # print('after set', self.__dict__)
                else:
                    raise TypeError(
                        f'Expect type for parameter {key} is {type_expected}, but got \'{value}\' which has type {type(value)} ')

        elif children:
            for child in children:
                child.recursive_set_attr(key, value)

    def get_surface_dict(self) -> dict:

        def expandable_component(_obj):
            expandable = [_ for _ in _obj.__dict__.values() if
                          isinstance(_, Parameters) or isinstance(_, Configuration)]

            return expandable

        def terminal_check(obj: object) -> bool:
            is_terminal = isinstance(obj, Parameters) and all((
                len(expandable_component(component)) == 0 for component in expandable_component(obj)
            ))
            return is_terminal

        def non_terminal_check(obj: object) -> bool:
            is_non_terminal = isinstance(obj, Parameters) or isinstance(obj, Configuration) and any((
                len(expandable_component(component)) != 0 for component in expandable_component(obj)
            ))
            return is_non_terminal

        children_nonterminal = filter(non_terminal_check, self.__dict__.values())
        children_terminal = filter(terminal_check, self.__dict__.values())

        children_nonterminal_dicts = [child.get_surface_dict() for child in children_nonterminal]
        children_terminal_dicts = [child.__dict__ for child in children_terminal]
        surface_dict = {}
        for d in children_nonterminal_dicts + children_terminal_dicts:
            surface_dict.update(d)
        return surface_dict


@dataclass
class Parameters(LispConvertable):

    def to_lisp_command(self) -> str:
        convert_dict = {True: 't', False: 'nil'}
        commands = []
        for key, value in self.__dict__.items():
            if value is not None:
                if isinstance(value, Parameters):
                    value = value.to_lisp_command()
                print_value = value
                if type(value) is bool:
                    print_value = convert_dict[value]
                sub_command = f':{key} {print_value}'
                commands.append(sub_command)
        lisp_command = ' '.join(commands)
        return lisp_command

    def show(self):
        print(self.__repr__())  # formatting


SingleViewpoint = Literal[
    'onset', 'cpitch', 'dur', 'keysig', 'mode', 'tempo', 'pulses', 'barlength', 'deltast', 'bioi', 'phrase',
    'mpitch', 'accidental', 'dyn', 'voice', 'ornament', 'comma', 'articulation',
    'ioi', 'posinbar', 'dur-ratio', 'referent', 'cpint', 'contour', 'cpitch-class', 'cpcint', 'cpintfref', 'cpintfip', 'cpintfiph', 'cpintfb', 'inscale',
    'ioi-ratio', 'ioi-contour', 'metaccent', 'bioi-contour', 'lphrase', 'cpint-size', 'newcontour', 'cpcint-size', 'cpcint-2', 'cpcint-3', 'cpcint-4', 'cpcint-5', 'cpcint-6', 'octave', 'tessitura', 'mpitch-class',
    'registral-direction', 'intervallic-difference', 'registral-return', 'proximity', 'closure'
]


@dataclass
class RequiredParameters(Parameters):
    dataset_id: str = None
    target_viewpoints: List[SingleViewpoint] = None
    source_viewpoints: Union[Literal[':select'],
                             List[Union[SingleViewpoint,
                                        Tuple[SingleViewpoint]]]] = None

    def is_complete(self) -> bool:
        condition = all([
            # type(self.dataset_path) is str,
            bool(self.target_viewpoints),
            bool(self.source_viewpoints)
        ])
        return condition

    def viewpoint_to_string(self, viewpoint: Union[SingleViewpoint, Tuple[SingleViewpoint]]) -> str:
        if type(viewpoint) is str:
            string = str(viewpoint)
        elif type(viewpoint) is tuple:
            string = self.tuple_to_command(viewpoint)
        else:
            raise TypeError(viewpoint)
        return string

    def list_to_command(self, l: Iterable) -> str:
        list_of_string = [self.viewpoint_to_string(viewpoint=viewpoint) for viewpoint in l]
        long_string = ' '.join(list_of_string)
        command = f'\'({long_string})'
        return command

    def tuple_to_command(self, t: tuple) -> str:
        list_of_string = [self.viewpoint_to_string(viewpoint=viewpoint) for viewpoint in t]
        long_string = ' '.join(list_of_string)
        command = f'({long_string})'
        return command

    def dataset_id_to_command(self):
        command_dataset_id = self.dataset_id
        return command_dataset_id

    def target_viewpoints_to_command(self):
        command_target_viewpoints = self.list_to_command(self.target_viewpoints)
        return command_target_viewpoints

    def source_viewpoints_to_command(self):
        if self.source_viewpoints == ':select':
            command_source_viewpoints = self.source_viewpoints
            # raise NotImplementedError
        elif type(self.source_viewpoints) is list:
            command_source_viewpoints = self.list_to_command(self.source_viewpoints)
        else:
            raise TypeError(self.source_viewpoints)
        return command_source_viewpoints

    def to_lisp_command(self) -> str:
        if not self.is_complete():
            raise AssertionError(f'{self} Missing required argument')
        command = f'{self.dataset_id_to_command()} {self.target_viewpoints_to_command()} {self.source_viewpoints_to_command()}'
        return command


@dataclass
class LTMOModelOptions(Parameters):
    ltmo_order_bound: int = None
    ltmo_mixtures: bool = None
    ltmo_update_exclusion: bool = None
    ltmo_escape: Literal[':a', ':b', ':c', ':d', ':x'] = None

    def to_lisp_command(self) -> str:
        generic_command = super().to_lisp_command()
        command = f'\'({generic_command})'
        command = command.replace('ltmo_', '')
        command = command.replace('_', '-')
        return command


@dataclass
class STMOModelOptions(Parameters):
    stmo_order_bound: int = None
    stmo_mixtures: bool = None
    stmo_update_exclusion: bool = None
    stmo_escape: Literal[':a', ':b', ':c', ':d', ':x'] = None

    def to_lisp_command(self) -> str:
        generic_command = super().to_lisp_command()
        command = f'\'({generic_command})'
        command = command.replace('stmo_', '')
        command = command.replace('_', '-')
        return command

@dataclass
class StatisticalModellingParameters(Parameters):
    models: Literal[':stm', ':ltm', ':ltm+', ':both', ':both+'] = None
    ltmo: Literal[':ltmo'] = None
    ltmo_options: LTMOModelOptions = field(default_factory=LTMOModelOptions)
    stmo: Literal[':stmo'] = None
    stmo_options: STMOModelOptions = field(default_factory=STMOModelOptions)

    def models_to_lisp_command(self) -> str:
        command = f':models {self.models}'
        return command

    def ltmo_to_lisp_command(self) -> str:
        command = self.ltmo
        return command

    def stmo_to_lisp_command(self) -> str:
        command = self.stmo
        return command

    def ltmo_options_to_lisp_command(self) -> str:
        command = self.ltmo_options.to_lisp_command()
        return command

    def stmo_options_to_lisp_command(self) -> str:
        command = self.stmo_options.to_lisp_command()
        return command

    def to_lisp_command(self) -> str:
        if self.models is not None and self.ltmo is None and self.stmo is None:
            command = f'{self.models_to_lisp_command()}'
            return command

        elif self.ltmo is not None and self.stmo is None:
            command = f'{self.models_to_lisp_command()} {self.ltmo_to_lisp_command()} {self.ltmo_options_to_lisp_command()}'
            return command

        elif self.stmo is not None and self.ltmo is None:
            command = f'{self.models_to_lisp_command()} {self.stmo_to_lisp_command()} {self.stmo_options_to_lisp_command()}'
            return command

        elif self.stmo is not None and self.ltmo is not None:
            command = f'{self.models_to_lisp_command()} {self.stmo_to_lisp_command()} {self.stmo_options_to_lisp_command()} {self.ltmo_to_lisp_command()} {self.ltmo_options_to_lisp_command()}'
            return command

        elif self.stmo is None and self.stmo_options is not None:
            raise ValueError(
                f'Please specify \':stmo\' if you want to further specify the STM optional properties {self.stmo_options_to_lisp_command()}')

        elif self.ltmo is None and self.ltmo_options is not None:
            raise ValueError(
                f'Please specify \':ltmo\' if you want to further specify the LTM optional properties {self.ltmo_options_to_lisp_command()}')


@dataclass
class TrainingParameters(Parameters):
    pretraining_id: str = None
    k: Union[int, Literal[":full"]] = 10
    resampling_indices: List[int] = None

    def pretraining_id_to_command(self) -> str:
        command = self.pretraining_id
        return command

    def k_to_command(self):
        command_k = f':k {self.k}'
        return command_k

    def resampling_indices_to_command(self):
        raise NotImplementedError

    def to_lisp_command(self) -> str:
        if self.pretraining_id:
            command = f':pretraining-ids \'({self.pretraining_id_to_command()}) {self.k_to_command()}'
            return command

        else:
            command = f'{self.k_to_command()}'
            return command


@dataclass
class BasisOption(Parameters):
    basis: Union[List[SingleViewpoint],
                 Literal[':pitch-full', ':pitch-short', ':bioi', ':onset', ':auto']] = None

    def to_lisp_command(self) -> str:
        generic_command = super().to_lisp_command()
        if generic_command:
            command = f'\'({generic_command})'
        else:
            command = ''
        return command


@dataclass
class ViewpointSelectionParameters(Parameters):
    """
    When the source viewpoint supplied is :select
    """

    basis: BasisOption = None
    dp: int = None
    max_links: int = None
    min_links: int = None
    viewpoint_selection_output: str = None  # a filepath to write output for every viewpoint system considered during viewpoint selection.
    # The default is nil meaning that no files are written.


@dataclass
class OutputParameters(Parameters):
    detail: Literal[1, 2, 3] = 3
    output_path: str = None
    overwrite: bool = False  # whether to overwrite an existing output file if it exists
    separator: str = None  # a string defining the character to use for delimiting columns in the output file

    def detail_to_command(self) -> str:
        command_detail = f':detail {self.detail}'
        return command_detail

    # the default output path in py2lispIDyOM is experiment_history/THIS_EXP/experiment_output_data_folder/
    def output_path_to_command(self) -> str:
        command_outpath = f':output-path \"{self.output_path}\"'
        return command_outpath

    def overwrite_to_command(self) -> str:
        convert_dict = {True: 't', False: 'nil'}
        if self.overwrite is True:
            command_overwrite = f':overwrite {convert_dict[True]}'
            return command_overwrite
        if self.overwrite is False:
            command_overwrite = f':overwrite {convert_dict[False]}'
            return command_overwrite

    def to_lisp_command(self) -> str:
        command = f'{self.detail_to_command()} {self.output_path_to_command()} {self.overwrite_to_command()}'
        return command


@dataclass
class CachingParameters(Parameters):
    use_resampling_set_cache: bool = None
    use_ltms_cache: bool = None


@dataclass
class ExperimentLogger:
    test_dataset_path: str
    pretrain_dataset_path: str
    experiment_history_folder_path: str

    def __post_init__(self):

        self.experiment_history_folder = self.generate_experiment_history_folder()
        self.this_exp_folder = self.generate_this_exp_folder()
        self.input_data_exp_folder = self.generate_input_data_exp_folder()
        self.test_dataset_exp_folder = self.generate_test_dataset_exp_folder()
        self.train_dataset_exp_folder = self.generate_pretrain_dataset_exp_folder()
        self.output_data_exp_folder = self.generate_output_data_exp_folder()

    def generate_experiment_history_folder(self):
        if self.experiment_history_folder_path is None:
            experiment_history_folder = 'experiment_history/'
            if not os.path.exists(experiment_history_folder):
                os.makedirs(experiment_history_folder)
        else:
            experiment_history_folder = self.experiment_history_folder_path
        return experiment_history_folder

    def generate_this_exp_folder(self):
        today_date = datetime.date.today()
        now_time = datetime.datetime.now()
        this_experiment_folder = self.experiment_history_folder + today_date.strftime(
            '%d-%m-%y') + '_' + now_time.strftime(
            '%H.%M.%S') + '/'
        os.makedirs(this_experiment_folder)
        return this_experiment_folder

    def generate_input_data_exp_folder(self):
        input_data_folder = self.this_exp_folder + 'experiment_input_data_folder/'
        os.makedirs(input_data_folder)
        return input_data_folder

    def _get_files_from_paths(self, path):  # only two types files allowed: midi and kern
        files = []
        for file in glob(path + '*'):
            if file[file.rfind("."):] == ".mid" or ".krn":
                files.append(file)
        return natsorted(files)

    def _put_midis_in_folder(self, files, folder_path):
        for file in files:
            shutil.copyfile(file, folder_path + file[file.rfind("/"):])

    def generate_test_dataset_exp_folder(self):
        input_data_folder = self.input_data_exp_folder
        test_folder = input_data_folder + 'test_dataset/'
        os.makedirs(test_folder)
        test_dataset_path = self.test_dataset_path
        test = self._get_files_from_paths(test_dataset_path)
        self._put_midis_in_folder(test, test_folder)
        if not os.listdir(test_folder):
            raise AssertionError(f'test_dataset folder is empty!')
        else:
            print("** Putting Test dataset files in experiment history folder. **")

        return test_folder

    def generate_pretrain_dataset_exp_folder(self):
        pretrain_dataset_path = self.pretrain_dataset_path
        if pretrain_dataset_path is None:
            pass
            print('** No pretraining dataset detected. **')
        else:
            input_data_folder = self.input_data_exp_folder
            pretrain_folder = input_data_folder + 'pretrain_dataset/'
            os.makedirs(pretrain_folder)
            train = self._get_files_from_paths(pretrain_dataset_path)
            self._put_midis_in_folder(train, pretrain_folder)
            if not os.listdir(pretrain_folder):
                raise AssertionError(f'pretrain_dataset folder is empty!')
            else:
                print("** Putting Pretraining dataset files in experiment history folder. **")
            return pretrain_folder


    # def put_test_midi_in_exp_folder(self):
    #     print("** Putting Test dataset files in experiment history folder. **")
    #     test_files = self._get_files_from_paths(self.test_dataset_path)
    #     self._put_midis_in_folder(test_files, self.test_dataset_exp_folder)

    # def put_train_midi_in_exp_folder(self):
    #     if self.pretrain_dataset_path is None:
    #         pass
    #         print('** No pretraining dataset detected. **')
    #     else:
    #         print("** Putting pretrain dataset files in experiment history folder. **")
    #         train = self._get_files_from_paths(self.pretrain_dataset_path)
    #         self._put_midis_in_folder(train, self.train_dataset_exp_folder)

    def generate_output_data_exp_folder(self):
        output_data_folder = self.this_exp_folder + 'experiment_output_data_folder/'
        os.makedirs(output_data_folder)
        return output_data_folder


@dataclass
class Configuration(LispConvertable, ABC):
    pass


@dataclass(repr=False)
class RunModelConfiguration(Configuration):
    this_exp_log_path: str = None
    test_dataset_id: str = None
    pretrain_dataset_id: str = None

    output_parameters: OutputParameters = field(default_factory=OutputParameters)
    required_parameters: RequiredParameters = field(default_factory=RequiredParameters)
    statistical_modelling_parameters: StatisticalModellingParameters = field(
        default_factory=StatisticalModellingParameters)
    training_parameters: TrainingParameters = field(default_factory=TrainingParameters)
    viewpoint_selection_parameters: ViewpointSelectionParameters = field(default_factory=ViewpointSelectionParameters)
    caching_parameters: CachingParameters = field(default_factory=CachingParameters)

    def to_lisp_command(self) -> str:
        # assert self.required_parameters._is_available(), self.required_parameters
        all_parameters = [
            self.required_parameters,
            self.statistical_modelling_parameters,
            self.training_parameters,
            self.viewpoint_selection_parameters,
            self.output_parameters,
            self.caching_parameters,
        ]
        subcommands = [parameters.to_lisp_command() for parameters in all_parameters]
        non_empty_subcommands = [x for x in subcommands if x != '']
        joined_commands = ' '.join(non_empty_subcommands)
        command = f'(idyom:idyom {joined_commands})'
        return command


@dataclass
class DatabaseConfiguration(Configuration):
    this_exp_log_path: str = None
    test_dataset_id: str = None
    pretrain_dataset_id: str = None
    test_dataset_Name: str = 'TEST_DATASET'
    pretrain_dataset_Name: str = 'PRETRAIN_DATASET'

    def to_lisp_command(self):
        if self.pretrain_dataset_id:
            commands = [
                self.get_command_import_testdb(),
                self.get_command_import_traindb(),
            ]
        else:
            commands = [
                self.get_command_import_testdb(),
            ]
        total_command = '\n'.join(commands)
        return total_command

    @property
    def test_dataset_log_path(self):
        test_dataset_log_path = self.this_exp_log_path + 'experiment_input_data_folder/test_dataset/'
        return test_dataset_log_path

    @property
    def train_dataset_log_path(self):
        train_dataset_log_path = self.this_exp_log_path + 'experiment_input_data_folder/pretrain_dataset/'
        return train_dataset_log_path

    def get_music_files_type(self, path) -> str:
        for file in glob(path + '*'):
            if file[file.rfind("."):] == ".mid":
                return str('mid')
            if file[file.rfind("."):] == ".krn":
                return str('krn')
            else:
                print('Music file type unsupported. Please use either midi files or kern files.')

    def _oneline_import_db_to_lisp_command(self, file_type, Path, Name, ID) -> str:
        subcommands = [f':{file_type}', f'\"{Path}\"', f'\"{Name}\"', ID]
        non_empty_subcommands = [x for x in subcommands if x != '']
        joined_commands = ' '.join(non_empty_subcommands)
        command = f'(idyom-db:import-data {joined_commands})'
        return command

    def _get_command_import_db(self, path, dataset_name, dataset_id):
        file_type = self.get_music_files_type(path)
        command_import_testdb = self._oneline_import_db_to_lisp_command(file_type=file_type,
                                                                        Path=path,
                                                                        Name=dataset_name,
                                                                        ID=dataset_id)
        return command_import_testdb

    def get_command_import_testdb(self):
        command = self._get_command_import_db(path=self.test_dataset_log_path, dataset_name=self.test_dataset_Name,
                                              dataset_id=self.test_dataset_id)
        return command

    def get_command_import_traindb(self):
        command = self._get_command_import_db(path=self.train_dataset_log_path, dataset_name=self.pretrain_dataset_Name,
                                              dataset_id=self.pretrain_dataset_id)
        return command

    def to_lisp_command_import_testdb(self) -> str:
        return self.get_command_import_testdb()

    def to_lisp_command_import_traindb(self) -> str:
        return self.get_command_import_traindb()

    # def to_lisp_command(self) -> str:
    #     commands = [
    #         self.get_command_import_testdb(),
    #         self.get_command_import_traindb(),
    #     ]
    #     total_command = '\n'.join(commands)
    #     return total_command


@dataclass
class IDyOMConfiguration(Configuration):
    database_configuration: DatabaseConfiguration = field(default_factory=DatabaseConfiguration)
    run_model_configuration: RunModelConfiguration = field(default_factory=RunModelConfiguration)

    def to_lisp_command(self) -> str:
        if self.run_model_configuration.training_parameters.pretraining_id:
            commands = [
                self.start_idyom_command(),
                self.import_test_dataset_command(),
                self.import_train_dataset_command(),
                self.run_model_command(),
                self.quit_command()
            ]
            total_command = '\n'.join(commands)
            return total_command
        else:
            commands = [
                self.start_idyom_command(),
                self.import_test_dataset_command(),
                self.run_model_command(),
                self.quit_command()
            ]
            total_command = '\n'.join(commands)
            return total_command

    def start_idyom_command(self) -> str:
        command = '(start-idyom)'
        return command

    def import_test_dataset_command(self) -> str:
        command = self.database_configuration.to_lisp_command_import_testdb()
        return command

    def import_train_dataset_command(self) -> str:
        command = self.database_configuration.to_lisp_command_import_traindb()
        return command

    def import_datasets_command(self) -> str:
        command = self.database_configuration.to_lisp_command()
        return command

    def describe_database_command(self) -> str:
        command = '(idyom-db:describe-database)'
        return command

    def describe_detailed_database_command(self) -> str:
        command = '(idyom-db:describe-database :verbose t)'
        return command

    def run_model_command(self) -> str:
        command = self.run_model_configuration.to_lisp_command()
        return command

    def quit_command(self) -> str:
        command = '(quit)'
        return command
